const generators: Record<string, Fig.Generator> = {
  servicesgenerators: {
    script: "brew services list | sed -e 's/ .*//' | tail -n +2",
    postProcess: function (out) {
      return out
        .split("\n")
        .filter((line) => !line.includes("unbound"))
        .map((line) => ({
          name: line,
          type: "option",
        }));
    },
  },
};

export const completionSpec: Fig.Spec = {
  name: "brew",
  description: "Package manager for macOS",
  subcommands: [
    { name: "list", description: "List all installed formulae" },
    {
      name: "leaves",
      description:
        "List installed formulae that are not dependencies of another installed formula",
      options: [
        {
          name: ["-r", "--installed-on-request"],
          description: "Show manually installed formula.",
        },
        {
          name: ["-p", "--installed-as-dependency"],
          description: "Show installed formula as dependencies.",
        },
      ],
    },
    {
      name: "doctor",
      description: "Check your system for potential problems",
      options: [
        {
          name: "--list-checks",
          description: "List all audit methods.",
        },
        {
          name: ["-D", "--audit-debug"],
          description: "Enable debugging and profiling of audit methods.",
        },
      ],
    },
    {
      name: "info",
      description: "Display brief statistics for your Homebrew installation",
    },
    {
      name: "update",
      description: "Fetch the newest version of Homebrew and all formulae",
    },
    {
      name: "outdated",
      description:
        "List installed casks and formulae that have an updated version available",
      options: [
        {
          name: ["-d", "--debug"],
          description: "Display any debugging information.",
        },
        {
          name: ["-q", "--quiet"],
          description: "List only the names of outdated kegs.",
        },
        {
          name: ["-v", "--verbose"],
          description: "Include detailed version information.",
        },
        {
          name: ["-h", "--help"],
          description: "Show help message for the outdated command.",
        },
        { name: "--cask", description: "List only outdated casks." },
        {
          name: "--fetch-HEAD",
          description:
            "Fetch the upstream repository to detect if the HEAD installation of the formula is outdated.",
        },
        { name: "--formula", description: "List only outdated formulae." },
        {
          name: "--greedy",
          description:
            "Print outdated casks with auto_updates or version :latest.",
        },
        { name: "--json", description: "Print output in JSON format." },
      ],
    },
    { name: "upgrade", description: "Upgrade outdated casks and outdated" },
    {
      name: "search",
      description:
        "Perform a substring search of cask tokens and formula names",
    },
    {
      name: "config",
      description: "Show Homebrew and system configuration info",
    },
    {
      name: "install",
      description: "Install <formula>",
      insertValue: "install ",
      args: {
        variadic: true,
        name: "formula",
        description: "Formula or cask to install",
        generators: {
          script:
            "HBPATH=$(brew --repository); ls -1 $HBPATH/Library/Taps/homebrew/homebrew-core/Formula $HBPATH/Library/Taps/homebrew/homebrew-cask/Casks",
          postProcess: function (out) {
            return out.split("\n").map((formula) => {
              return {
                name: formula.replace(".rb", ""),
                description: "formula",
                icon: "🍺",
                priority:
                  (formula[0] >= "0" && formula[0] <= "9") || formula[0] == "/"
                    ? 0
                    : 51,
              };
            });
          },
        },
      },
    },
    {
      name: "uninstall",
      description: "Uninstall <formula>",
      args: {
        variadic: true,
        name: "formula",
        generators: {
          script: "brew list -1 --formulae",
          postProcess: function (out) {
            return out.split("\n").map((formula) => {
              return {
                name: formula,
                icon: "🍺",
                description: "Installed formula",
              };
            });
          },
        },
      },
    },
    {
      name: "cask",
      insertValue: "cask ",
      description:
        "Homebrew Cask provides a friendly CLI workflow for the administration of macOS applications distributed as binaries.",
      subcommands: [
        {
          name: "install",
          insertValue: "install ",
          description: "Installs the given cask",
          args: {
            name: "cask",
            description: "Cask to install",
          },
        },
        {
          name: "uninstall",
          insertValue: "uninstall ",
          description: "Uninstalls the given cask",
          args: {
            variadic: true,
            generators: {
              script: "brew list -1 --cask",
              postProcess: function (out) {
                return out.split("\n").map((formula) => {
                  return {
                    name: formula,
                    icon: "🍺",
                    description: "Installed formula",
                  };
                });
              },
            },
          },
        },
      ],
    },
    {
      name: "cleanup",
      insertValue: "cleanup ",
      description:
        "Remove stale lock files and outdated downloads for all formulae and casks and remove old versions of installed formulae.",
      options: [
        {
          name: ["--prune", "--prune=all"],
          description: "Remove all cache files older than specified days.",
        },
        {
          name: ["-n", "--dry-run"],
          description:
            "Show what would be removed, but do not actually remove anything.",
        },
        {
          name: "-s",
          description:
            "Scrub the cache, including downloads for even the latest versions.",
        },
        {
          name: "--prune-prefix",
          description:
            "Only prune the symlinks and directories from the prefix and remove no other files.",
        },
      ],
      args: {
        variadic: true,
        generators: generators.servicesGenerator,
      },
    },
    {
      name: "services",
      description:
        "Manage background services with macOS' launchctl(1) daemon manager.",
      options: [
        {
          name: ["-d", "--debug"],
          description: "Display any debugging information.",
        },
        {
          name: ["-q", "--quiet"],
          description: "Suppress any warnings.",
        },
        {
          name: ["-v", "--verbose"],
          description: "Make some output more verbose.",
        },
        {
          name: ["-h", "--help"],
          description: "Get help with services command",
        },
      ],
      subcommands: [
        {
          name: "cleanup",
          insertValue: "cleanup",
          description: "Remove all unused services.",
        },
        {
          name: "list",
          insertValue: "list",
          description: "List all services.",
        },
        {
          name: "run",
          insertValue: "run ",
          description:
            "Run the service formula without registering to launch at login (or boot).",
          options: [
            {
              name: "--all",
              insertValue: "--all",
              description: "Start all services",
            },
          ],
          args: {
            variadic: true,
            generators: generators.servicesGenerator,
          },
        },
        {
          name: "start",
          insertValue: "start ",
          description:
            "Start the service formula immediately and register it to launch at login",
          options: [
            {
              name: "--all",
              insertValue: "--all",
              description: "Start all services",
            },
          ],
          args: {
            variadic: true,
            generators: generators.servicesGenerator,
          },
        },
        {
          name: "stop",
          insertValue: "stop ",
          description:
            "Stop the service formula immediately and unregister it from launching at",
          options: [
            {
              name: "--all",
              insertValue: "--all",
              description: "Start all services",
            },
          ],
          args: {
            variadic: true,
            generators: generators.servicesGenerator,
          },
        },
        {
          name: "restart",
          insertValue: "restart ",
          description:
            "Stop (if necessary) and start the service formula immediately and register it to launch at login (or boot).",
          options: [
            {
              name: "--all",
              insertValue: "--all",
              description: "Start all services",
            },
          ],
          args: {
            variadic: true,
            generators: generators.servicesGenerator,
          },
        },
      ],
    },
    {
      name: "analytics",
      description: "Manages analytics preferences",
      subcommands: [
        {
          name: "on",
          description: "Turns on analytics",
        },
        {
          name: "off",
          description: "Turns off analytics",
        },
        {
          name: "regenerate-uuid",
          description: "Regenerate the UUID used for analytics",
        },
      ],
    },
  ],
  options: [
    {
      name: ["--version"],
      description: "The current Homebrew version",
    },
  ],
};
